dl.field-list .doc-param-default, dl.doc-field-list .doc-param-default {
  float: none;
}

.field-list > dl, dl.field-list, dl.doc-field-list {
  display: flex;
  flex-flow: row wrap;
  padding-left: 10px;
}

.field-list > dl > dt, dl.field-list > dt, dl.doc-field-list > dt {
  flex-basis: 20%;
  font-weight: bold;
  word-break: break-word;
  padding: 10px 0;
  border-bottom: 1px solid #e5e5e5;
}

.field-list > dl > dt:after, dl.field-list > dt:after {
  content: ":";
}

[dir=ltr] .field-list > dl > dd, dl.field-list > dd.doc-field-def, dl.doc-field-list > dd.doc-field-def {
  flex-basis: 70%;
  flex-grow: 1;
  margin: 0;
  padding: 10px 0 10px 10px;
  border-bottom: 1px solid #e5e5e5;
}

dd.doc-field-def > p:last-child {
  padding-bottom: 0;
  margin-bottom: 0;
}
